DEBUG	EQU	0
;
;	SCSIStop.asm: Kill off everything.
;	This program is in the public domain.
;		Ephraim M. Vishniac, October 5th, 1986.
;
;	The current version of the Finder (5.3) doesn't make any
;	provision for notifying device drivers when the system is
;	being Shut Down.  This means that there's no convenient 
;	way to do an orderly shutdown of attached peripheral
;	devices.
;	The Finder does the following when shutting down:
;		1. Deliver a "goodbye kiss" to each device driver
;		   that needs it.
;		2. Ejects and Unmounts every volume.
;		3. Resets.
;	It would make sense for the Finder to close every device
;	driver between (2) and (3), but it doesn't.  Apple Technical
;	Support tells me that "Future system software will have a
;	mechanism for shutdown notification ('future' means 'sometime
;	next year', most likely)."  They don't say what the 
;	mechanism will be.
 
;	This program does everything that the Finder's
;	Shut Down choice does, plus send a Stop command to each
;	SCSI device.  The effect of Stop varies from one SCSI
;	device to another.  Some disks "park" the disk heads;
;	some actually spin down the disk.  Consult your owner's
;	manual or technical reference.

;	Please direct questions or comments to the author:

;	uucp:		decvax!wanginst!wang!ephraim

;	MCI mail:	masstech

;	US mail:	Ephraim Vishniac
;			P.O. Box 1357
;			East Arlington, MA 02174


;	Includes:

	Include MacTraps.D	; Use System and ToolBox traps
	Include SysEqu.D	; Use System equates
	Include	FSEqu.D		; Use file system equates
	Include	ToolEqu.D	; Use toolbox equates
	Include	SysErr.D	; Use system error equates
	Include	SCSIEqu.Txt	; Use SCSI-specific equates & macros

;	Macros:

MACRO	DoAlert		AlertNumber =
	; Alert (AlertNumber:INT, NIL:ProcPtr) : INT;
	Clr.w		-(sp)			; space for result
	move.w		#{AlertNumber},-(sp)	; alert number
	Clr.l		-(sp)			; NIL filter
	_Alert
	move.w		(sp)+,d0		; d0 = button hit
	cmp.w		#OKitem,d0		; standard test
	|

;	Literals:

Greeting	equ	128		; Greeting alert number
OKitem		equ	1		; Item number for OK button
QuitItem	equ	2		; Item number for Quit button

StopSize	equ	6		; Size of stop command

Second		equ	60		; number of ticks in a second
StopTime	equ	10*Second	; we're in no hurry

;	Global data:

StatWord	ds.w	1	; Status word for _SCSIComplete
MsgWord		ds.w	1	; Message word for _SCSIComplete

Start

;------------------------------- InitManagers -----------------------------
;	This init code is from one of Apple's example programs.

	pea		-4(A5)			; Quickdraw's global area
	_InitGraf				; Init Quickdraw
	_InitFonts				; Init Font Manager
	move.l		#$FFFF,D0		; Flush all events
	_FlushEvents
	_InitWindows				; Init Window Manager
	_InitMenus				; Init Menu Manager
	_TEInit					; Init Text Edit
	Clr.l		-(SP)			; No restart procedure
	_InitDialogs				; Init Dialog Manager
	_InitCursor				; Turn on arrow cursor
;--------------------------------------------------------------------------

;	Warn the user that this will end it all.

	DoAlert		Greeting
	beq		DropDead

;	The user didn't say OK.  Go back to the Finder.

	_ExitToShell

;	The user said it's OK, so let's go.

DropDead
	IF	DEBUG
	_Debugger
	ENDIF

	; Detach our code segment for safety.
	; Remember that files get closed as volumes are dismounted.
	Clr.l		-(sp)		; space for handle
	move.l		#'CODE',-(sp)	; resource type
	move.w		#1,-(sp)	; resource number
	_GetResource			; get our handle
	_DetachResource			; break it off

	; Say goodbye to each driver
	; This routine is digested from Finder 5.3.
SayGoodBye
	suba.w	#ioFQElSize,sp		; space for I/O parm block
	move.w	#-1,csCode(sp)		; goodbye code
	movea.l	UTableBase,A1		; A1 = Unit Table pointer
	move.w	UnitNtryCnt,D1		; D1 = Unit Table size
@0	move.l	(A1)+,D0		; D0 = DCE handle
	beq.s	@1			; skip if empty slot
	movea.l	D0,A0			; A0 = DCE handle
	movea.l	(A0),A0			; A0 = DCE pointer
	bTst	#dNeedGoodbye,dCtlFlags(A0)	; Need goodbye?
	beq.s	@1			; skip if no goodbye needed
	move.w	dCtlRefNum(A0),ioRefNum(sp)	; set driver refnum
	movea.l	sp,A0			; A0 = IOPB pointer
	_Control			; say goodbye
@1	subq.w	#1,D1			; that's one down
	bne.s	@0			; loop through table

	; Eject all volumes
	; This routine is severely digested from Finder 5.3.
EjectAll
	move.l	vcbQHdr+qHead,D7	; D7 = head of VCB queue
	move.l	sp,A0			; A0 = I/O parm block pointer
	Clr.l	ioFileName(A0)		; clear file name pointer
	bra.s	@1			; hop into ejection loop
@0	move.l	(A4),D7			; D7 = next VCB pointer
	move.w	vcbVRefNum(A4),D0	; D0 = Volume RefNum
	move.w	D0,ioVRefNum(A0)	; Set VRefNum in I/O parm block
	_Eject				; Stick it out
	_UnMountVol			; ditch it
@1	movea.l	D7,A4			; A4 = VCB pointer
	Tst.l	D7			; nil pointer
	bne.s	@0			; loop if not

	; Stop each SCSI device
	; This is new code.
SCSIStop
	move.w	#6,D7			; top SCSI address
@0	move.l	#StopTime,D0		; completion time for Stop
	move.w	#StopSize,D1		; length of stop command
	move.w	D7,D2			; SCSI address
	lea	StopCommand,A0		; Stop command
	jsr	SCSISimple		; Issue the stop command

	dbra	D7,@0			; Kill off each device

	; Drop the bomb
	; This is from Finder 5.3.
ResetMac
	movea.l	RomBase,A0
	Reset
	jmp	$A(A0)

;_______________________________________________________________________
;
; Routine:	SCSISimple - Issue SCSI command with no data transfer
;		This routine is very distantly derived from 'SCSICommon'
;		in Apple's sample SCSI driver.
; Arguments:
;	(A0)	(input)  -- SCSI command block pointer
;	D0.L	(input)  -- ticks we're willing to wait for completion
;	D1.W	(input)  -- SCSI command block size
;	D2.W	(input)  -- SCSI address
;  StatWord(A5)	(output) -- returned status from _SCSIComplete, if no err.
;   MsgWord(A5)	(output) -- returned Message from _SCSIComplete, if no err.
;	D0.W	(output) -- result code (error if non-0)
;
;_______________________________________________________________________

SCSISimple
	MoveM.L	D4-D6/A4,-(SP)		; save regs
	Move.L	D0,D4			; D4 = tick count
	Move.W	D1,D5			; D5 = command block size
	Move.W	D2,D6			; D6 = SCSI ID
	Move.L	A0,A4			; A4 = command block pointer

	Clr.W	-(SP)			; space for result code
	
	; Get the SCSI bus
	_SCSIGet			; "Hello?"
	Tst.W	(SP)			; (did it work?)
	Bne	TargetErr		; (no... give up)
	
	; Select our target
	Move.W	D6,-(SP)		; SCSI address
	_SCSISelect			; "Is there anybody out there?"
	Tst.W	(SP)			; (did it work?)
	Bne.S	TargetErr		; (no... give up)
	
	; Issue the SCSI command
	Move.L	A4,-(SP)		; SCSI command pointer
	Move.W	D5,-(SP)		; command length
	_SCSICmd			; do the command

	; Complete the SCSI command.
	Pea	StatWord(A5)		; get the status
	Pea	MsgWord(A5)		; and message words
	Move.L	D4,-(SP)		; tick count we're willing to wait
	_SCSIComplete			; go get 'em
	Tst.W	(SP)			; (did it work?)
	Bne.S	TargetErr		; (no. Give up.)
	
	;What were the Status bytes?
@3	Move.B	StatWord+1(A5),D0	; was there an SCSI error?
	Beq.S	SCSIDone		; no error, just exit.

;We got an error in contacting the target, or in SCSIComplete.
TargetErr
	Move.W	#ioErr,(SP)		; if we can't get bus or target.
	
;Exit, stage left.
SCSIDone
	Move.W	(SP)+,D0		; pop error code
	MoveM.L	(SP)+,D4-D6/A4		; restore regs
	Tst.W	D0			; test error code
	Rts				; and return

;	Static data

StopCommand
	dc.b	$1B		; Start/stop command = $1B
	dc.b	$01		; LUN = 0; Reserved = 0; Immed. = 1
	dc.b	$00		; Reserved = 0
	dc.b	$00		; Reserved = 0
	dc.b	$00		; Reserved = 0; Start = 0
	dc.b	$00		; Reserved = 0; Flag = 0; Link = 0

	END
